@@ -1,7 +1,7 @@
 // repotool queries and manipulate multiple repository types in a uniform way.
 package main
 
-// Copyright by Eric S. Raymond
+// SPDX-FileCopyrightText: Copyright Eric S. Raymond <esr@thyrsus.com>
 // SPDX-License-Identifier: BSD-2-Clause
 
 import (
@@ -67,7 +67,7 @@ var makefileTemplate = `# Makefile for {{.Project}} conversion using reposurgeon
 # 1. Make sure reposurgeon and repotool are on your $PATH.
 # 2. (Skip this step if you're starting from a stream file.) For svn, set
 #    REMOTE_URL to point at the remote repository you want to convert;
-#    you can use either an svn: URL or an rsync: URL for this.
+#    you can use either an svn: URL, an rsync: URL, or a p4: URL for this.
 #    If the repository is already in a DVCS such as hg or git,
 #    set REMOTE_URL to either the normal cloning URL (starting with hg://,
 #    git://, etc.) or to the path of a local clone.
@@ -98,8 +98,7 @@ var makefileTemplate = `# Makefile for {{.Project}} conversion using reposurgeon
 
 EXTRAS = 
 REMOTE_URL = svn://svn.debian.org/{{.Project}}
-#REMOTE_URL = https://{{.Project}}.googlecode.com/svn/
-CVS_HOST = {{.Project}}.cvs.sourceforge.net
+CVS_HOST = cvs.sourceforge.net
 #CVS_HOST = cvs.savannah.gnu.org
 CVS_MODULE = {{.Project}}
 #REMOTE_URL = cvs://$(CVS_HOST)/{{.Project}}\#$(CVS_MODULE)
@@ -122,7 +121,7 @@ default: {{.Project}}-{{.TargetVCS}}
 
 # Build the repository from the stream dump
 {{.Project}}-{{.TargetVCS}}: {{.Project}}.{{.SourceVCS}} {{.Project}}.opts {{.Project}}.lift {{.Project}}.map $(EXTRAS)
-	$(REPOSURGEON) $(VERBOSITY) 'logfile $(LOGFILE)' 'script {{.Project}}.opts' "read $(READ_OPTIONS) <{{.Project}}.{{.SourceVCS}}" 'authors read <{{.Project}}.map' 'sourcetype {{.SourceVCS}}' 'prefer git' 'script {{.Project}}.lift' 'legacy write >{{.Project}}.fo' 'rebuild {{.Project}}-{{.TargetVCS}}'
+	$(REPOSURGEON) $(VERBOSITY) 'set logfile $(LOGFILE)' 'do {{.Project}}.opts' "read $(READ_OPTIONS) <{{.Project}}.{{.SourceVCS}}" 'authors read <{{.Project}}.map' 'select {{.SourceVCS}}' 'prefer git' 'do {{.Project}}.lift' 'legacy write >{{.Project}}.fo' 'rebuild {{.Project}}-{{.TargetVCS}}'
 
 # Build a stream dump from the local mirror
 {{.Project}}.{{.SourceVCS}}: {{.Project}}-mirror
@@ -177,7 +176,10 @@ gitk: {{.Project}}-git
 # This repack call is the active part of gc --aggressive.  This call is
 # tuned for very large repositories.
 gc: {{.Project}}-git
-	cd {{.Project}}-git; time git -c pack.threads=1 repack -AdF --window=1250 --depth=250
+	@cd {{.Project}}-git;
+	@echo "Before: $$(du -s -h .)"
+	@time git -c pack.threads=1 repack -AdF --window=1250 --depth=250
+	@echo "After: $$(du -s -h .)"
 `
 
 var acceptMissing bool
@@ -192,6 +194,7 @@ var progress bool
 
 var branch string
 var comparemode string
+var message string
 var refexclude string
 var revision string
 var basedir string
@@ -325,7 +328,7 @@ func makeStub(name string, contents string) {
 	fp.WriteString(contents)
 }
 
-func initialize(args []string) {
+func initmake(args []string) {
 	WriteSupport := newStringSet()
 	for _, vcs := range vcstypes {
 		if vcs.importer != "" {
@@ -341,11 +344,11 @@ func initialize(args []string) {
 	// Hacky special case implemented through extractor class
 	ReadSupport.Add("hg")
 	if verbose {
-		fmt.Printf("initialize args: %v\n", args)
+		fmt.Printf("initmake args: %v\n", args)
 	}
 	var squishy squishyParts
 	if len(args) < 1 {
-		croak("initialize requires a project name.")
+		croak("initmake requires a project name.")
 	}
 	project, args := args[0], args[1:]
 	squishy.Project = project
@@ -523,6 +526,22 @@ func mirror(args []string) {
 			}
 		})
 		baton.endProgress()
+	} else if strings.HasPrefix("p4://", operand) {
+		// According to
+		// https://stackoverflow.com/questions/31051732/perforce-to-git-migration,
+		// the clone operation will create a live Git repository capturing the
+		// Perforce history.
+		operand = operand[3:]
+		cmd := ""
+		legend := ""
+		if isdir(mirrordir) {
+			cmd = "git p4 sync --import-labels --detect-branches --destination %s %s@all"
+			legend = "p4 syncing"
+		} else {
+			cmd = fmt.Sprintf("git p4 clone --import-labels --detect-branches --destination %s %s@all")
+			legend = "p4 cloning"
+		}
+		runShellProcessOrDie(fmt.Sprintf(cmd, mirrordir, operand), legend)
 	} else if isdir(filepath.Join(operand, "locks")) {
 		if operand[0] == os.PathSeparator {
 			locald = operand
@@ -1171,10 +1190,10 @@ var helpdict = map[string]struct {
 	oneliner string
 	text     string
 }{
-	"initialize": {
-		"initialize",
+	"initmake": {
+		"initmake",
 		"create Makefile and stub files for standard conversion workflow",
-		`The "initialize" option takes a project name (and, optionally,
+		`The "initmake" option takes a project name (and, optionally,
 following source and target VCS types) and generates a
 Makefile that will sequence various steps of a repository
 conversion. It also generates stub lift and options files. This is
@@ -1194,10 +1213,10 @@ as Subversion dump files; the point is to be a lossless
 representation, or as close to one as possible.
 `},
 	"mirror": {
-		"mirror [URL] localdir",
+		"mirror {URL|localdir} [localname]",
 		"create or update a mirror of the source repository",
 		`The 'mirror' action makes or updates a local mirror of a
-Subversion, CVS, git, or hg repo. It requires a single argument,
+Subversion, CVS, git, hg, or p4 repo. It requires an argument,
 either a repository URL or the name of a local mirror directory
 created by a previous run. The first form creates a local mirror of
 the repository in a directory named after the last segment of the URL,
@@ -1211,21 +1230,29 @@ mirroring.
 Subversion URLs are as specified in the public documentation for
 Subversion.  CVS URLs must specify a host and repository path,
 followed by a '#', followed by a module name.  URLs for git and hg
-should be in the form normally used for clone commands.
-Alternatively, a repository URL may be a "file://" URL, in which
+should be in the form normally used for clone commands.  Perforce
+(p4) URLs must specify host, path, and depot (project).
+
+The mirror command can also be passed an rsync URL (service prefix
+"rsync://").  This will usually be faster than mirroring through an
+equivalent Subversion URL.
+
+Finally, a repository URL may be a "file://" URL, in which
 case the repository type is autodetected from the contents of
 the indicated directory. Note: A Subversion file URL has *three*
 slashes after the "file:" prefix!
-
-The mirror command can also be passed an rsync URL.  This
-will usually be faster than mirroring through an equivalent
-Subversion URL.
 `},
 	"branches": {
 		"branches",
 		"list repository branch names",
-		`The 'branches' option, run from within a repository directory ,
+		`The 'branches' action, run from within a repository directory ,
 returns a list of the repository's branch names.
+`},
+	"tags": {
+		"branches",
+		"list repository tag names",
+		`The 'tags' action, run from within a repository directory ,
+returns a list of the repository's tag names.
 `},
 	"checkout": {
 		"checkout [-r rev] [-t tag] [-b branch] [-o option]",
@@ -1296,7 +1323,7 @@ With a following argument that is a command name, display detailed help for that
 }
 
 var narrativeOrder []string = []string{
-	"initialize",
+	"initmake",
 	"export",
 	"mirror",
 	"branches",
@@ -1305,6 +1332,7 @@ var narrativeOrder []string = []string{
 	"compare-tags",
 	"compare-branches",
 	"compare-all",
+	"tags",
 	"version",
 	"help",
 }
@@ -1338,6 +1366,7 @@ func main() {
 	flags.StringVar(&branch, "b", "", "select branch for checkout or comparison")
 	flags.StringVar(&basedir, "d", "", "chdir to the argument repository path before doing checkout")
 	flags.StringVar(&refexclude, "e", "", "exclude pattern for tag and branch names.")
+	flags.StringVar(&message, "m", "", "argument is commit message text")
 	flags.StringVar(&revision, "r", "", "select revision for checkout or comparison")
 	flags.StringVar(&tag, "t", "", "select tag for checkout or comparison")
 	flags.StringVar(&passthrough, "o", "", "option passthrough")
@@ -1371,8 +1400,8 @@ func main() {
 	}
 
 	args := flags.Args()
-	if operation == "initialize" {
-		initialize(args)
+	if operation == "initmake" {
+		initmake(args)
 	} else if operation == "export" {
 		export()
 	} else if operation == "mirror" {
